function [VBparams, VBparamsPerIter] = stochasticCVB0(documents, numWords, numTopics, numDocumentIterations, dict, burnInPerDoc, minibatchSize, alpha, eta, saveParams, saveEvery, tau, kappa, tau2, kappa2, scale, initParams)
%Stochastic CVB0 for LDA.  This implementation is straightforward and very slow but
%should be useful as a reference implementation and for understanding the algorithm.
%It uses minibatches, but does not operate on sparse counts (clumping).
%Parameters 5 and above are optional.
%Reference: J. R. Foulds, L. Boyles, C. DuBois, P. Smyth and M. Welling. Stochastic collapsed variational Bayesian inference for latent Dirichlet allocation. Proceedings of the 19th ACM SIGKDD International Conference on Knowledge Discovery and Data Mining (KDD), 2013. 
%
%
%      To run a demo, try:
%      load NIPS.mat;
%      stochasticCVB0(documents, numWords, 50, 30000, dictionary);
%
%
%@author Jimmy Foulds. http://www.ics.uci.edu/~jfoulds/
%@requires: document order is random, token order is random.
%@param documents is a struct array, where documents(i).wordVector is an array containing the word indices for each token.
%@param numWords, numTopics. size of topic matrix.
%@param numDocumentIterations number of documents to process.
%@param dict a containers.Map object that maps word indices to words, for debug output of the top words in each topic.
%       Set it instead to [] to skip this output.
%@param burnInPerDoc number of burn-in passes per doc to learn theta.
%@param minibatchSize number of documents in minibatch.  Set to one to not use minibatches.
%@param alpha, eta. Dirichlet prior concentration parameters.
%@param saveParams a boolean determining whether per-iteration results are stored in VBparamsPerIter.
%@param saveEvery if saveParams, how many document iterations to wait between recorded results.
%recorded in VBparamsPerIter, as well as the final result in VBparams.
%@param tau, kappa, scale. Step size parameters for topics.
%@param tau2, kappa2. Step size parameters for theta.
%@param initParams VBparams from another run, to initialize at a warm start.
%@return VBparams A struct with the expected counts. add beta and normalize
%                 to recover the topics.
%@return VBparamsPerIter The VBparams variable, during training iterations.
%                        This is only returned if saveParams == true.

    debugTopics = true; %whether to display the topics during training
    
    %%%%%%%%%%% Fill in missing optional arguments %%%%%%%%%%%
    if nargin < 5
        dict = [];
    end
    if isempty(dict)
        debugTopics = false;
    end
    
    if nargin < 6
       burnInPerDoc = 1;
    end

    if nargin < 7
       minibatchSize = 10;
    end
    if nargin < 8
        alpha = ones(1,numTopics).* 0.1; % a row vector
    else
        alpha = ones(1,numTopics) .* alpha;
    end
    if nargin < 9
        eta = ones(numWords,1).*0.01;
    else
        eta = ones(numWords,1).*eta;
    end    
    
    if nargin < 10
        saveParams = false;
    end
    
    if nargin < 11
        saveEvery = length(documents);
    end
    
    %%% Step size parameters %%%
    if nargin < 12
        tau = 5;
    end
    if nargin < 13
        kappa = 0.9;
    end
    
    if nargin < 14
        tau2 = 5;
    end
    if nargin < 15
        kappa2 = 0.9;
    end
    if nargin < 16
        scale = 1;
    end

    %%%%%%%%%%% Initalization %%%%%%%%%%%     
    tic();
    if nargin < 17
        VBparams = initialize(documents, alpha, eta, numWords);
    else
        fprintf('Initializing to given params!\n');
        VBparams = initParams;
    end
    
    if saveParams
        VBparamsPerIter(floor(numDocumentIterations ./ saveEvery)) = VBparams;
        saveInd = 1;
    else
        VBparamsPerIter = [];
    end
    
    
    %The "_hat" variables hold results for the current minibatch
    wordTopicCounts_hat = zeros(numWords,numTopics);
    topicCounts_hat = zeros(1, numTopics);
    
    totalWordsInCorpus = 0;
    for i = 1:length(documents)
        totalWordsInCorpus = totalWordsInCorpus + length(documents(i).wordVector);
    end
    miniBatchesPerCorpus = length(documents)  ./ minibatchSize;
    VBparams.iter = 1;
    stepSize = scale ./tau^kappa;
    
    %%%%%%%%%%% Main VB loop %%%%%%%%%%%
    for numDocsSoFar = 1:numDocumentIterations        
        VBparams.iter = numDocsSoFar;
        if mod(numDocsSoFar, 100) == 0
            fprintf('Document iteration %d.\n', numDocsSoFar);
        end
        docInd = mod(numDocsSoFar -1, length(documents)) + 1;
        docLength = length(documents(docInd).wordVector);
        if docLength == 0
            continue;
        end
        
        %Burn in
        for burn=1:burnInPerDoc
            for wordInd = 1:docLength
                word = documents(docInd).wordVector(wordInd);
                stepSize2 = (wordInd + (burn-1) .* docLength + tau2)^-kappa2;

                %CVB0 estimate of gamma for current token
                probs = (VBparams.wordTopicCounts(word, :) + VBparams.eta(word)) ...
                     .*(VBparams.documentTopicCounts(docInd,:) + VBparams.alpha) ...
                     ./(VBparams.topicCounts + VBparams.etaSum);

                probs = probs ./ sum(probs);
                
                %update document statistics
                VBparams.documentTopicCounts(docInd,:) = (1 - stepSize2) * VBparams.documentTopicCounts(docInd,:) + stepSize2 .* docLength .* probs;
            end
        end
        
        %process the document, a word token at a time
        for wordInd = 1:length(documents(docInd).wordVector)
            word = documents(docInd).wordVector(wordInd);
            stepSize2 = (wordInd + burnInPerDoc .* docLength + tau2)^-kappa2;
            
            %%%%%%%%%% Main update code %%%%%%%%%%
            %CVB0 estimate of gamma for current token
            probs = (VBparams.wordTopicCounts(word, :) + VBparams.eta(word)) ...
                 .*(VBparams.documentTopicCounts(docInd,:) + VBparams.alpha) ...
                 ./(VBparams.topicCounts + VBparams.etaSum);

            probs = probs ./ sum(probs);

            %update document statistics
            VBparams.documentTopicCounts(docInd,:) = (1 - stepSize2) * VBparams.documentTopicCounts(docInd,:) + stepSize2 .* docLength .* probs;
            
            %update minibatch counts for topic statistics
            wordTopicCounts_hat(word, :) = wordTopicCounts_hat(word, :) + probs;
            topicCounts_hat = topicCounts_hat + probs;
            %%%%%%%%%% End main update code %%%%%%%%%%

        end
        
        if mod(numDocsSoFar, minibatchSize) == 0
            %Compensate for the minibatch size when choosing the step size.
            %This will make changing the minibatch size not require
            %choosing completely different step size parameters.
            stepSize = 1 - (1-stepSize)^minibatchSize;
            
            %Do a minibatch update
            VBparams.wordTopicCounts = (1 - stepSize) .* VBparams.wordTopicCounts + stepSize .* miniBatchesPerCorpus .* wordTopicCounts_hat;
            VBparams.topicCounts = (1 - stepSize) .* VBparams.topicCounts + stepSize .* miniBatchesPerCorpus .* topicCounts_hat;
            wordTopicCounts_hat = wordTopicCounts_hat .* 0;
            topicCounts_hat = topicCounts_hat .* 0;
            
            stepSize = scale ./(numDocsSoFar + tau)^kappa;
        end
        
        VBparams.wallClockTime = toc();
        
        %%%%%%%%%% Display the topics %%%%%%%%%%
        if debugTopics
            if mod(numDocsSoFar, 1000) == 0
                for t = 1:numTopics
                    topWords = getImportantWordsInTopic(VBparams.wordTopicCounts(:,t), dict, 10)
                end
            end
        end
        
        %optionally save results so far
        if saveParams && mod(numDocsSoFar, saveEvery) == 0
            VBparamsPerIter(saveInd) = VBparams;
            saveInd = saveInd + 1;
        end
    end
end

function VBparams = initialize(documents, alpha, eta, numWords)
%Initialize data structure for the model state
    numDocuments = length(documents);
    numTopics = length(alpha);
    
    VBparams.alpha = alpha;
    VBparams.eta = eta;
    VBparams.etaSum = sum(eta);
    VPparams.numTopics = numTopics;
    
    %pick a random document and initialize total counts based on its length
    numTokens = length(documents) .* length(documents(randi(numDocuments)).wordVector);
    VBparams.topicCounts = ones(1,numTopics) .* numTokens ./ numTopics;
    VBparams.wordTopicCounts = ones(numWords, numTopics) .* numTokens ./ (numWords .* numTopics);
    VBparams.iter = 0;
    
    %initialize document distributions
    VBparams.documentTopicCounts = zeros(length(documents), numTopics);
    for i = 1:length(documents)
        docLength = length(documents(i).wordVector);
        documentTopicCounts = rand(1, numTopics);
        VBparams.documentTopicCounts(i,:) = (documentTopicCounts ./ sum(documentTopicCounts)) * docLength;
    end
    VBparams.wallClockTime = 0;
end

function [topWords, weights] = getImportantWordsInTopic( topic, dictionary, numToGet)
%gets the words in the topic with the heighest weights, and outputs the
%actual words in them and their associated weights
%dictionary is a containers.Map object that maps word indices to words
    topWords = cell(numToGet, 1);
    weights = zeros(numToGet, 1);
    [topic, inds] = sort(topic, 'descend');
    for i = 1:numToGet
        if topic(i) == 0
            topWords = topWords(1:i-1);
            weights = weights(1:i-1);
            break;
        end
        topWords{i} = dictionary(inds(i));
        weights(i) = topic(i);
    end
end